#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest

from pyxel.calibration.fitness import (
    reduced_chi_squared,
    sum_of_abs_residuals,
    sum_of_squared_residuals,
)


@pytest.mark.parametrize(
    "simulated, target, weight, exp_result",
    [
        pytest.param(
            np.array([1, 2, 3, 4], dtype=float),
            np.array([1.1, 1.8, 3.3, 4], dtype=float),
            1,
            0.6,
            id="normal",
        ),
        pytest.param(
            np.array([1, 2, 3, 4, np.nan], dtype=float),
            np.array([1.1, 1.8, 3.3, 4, 5], dtype=float),
            1,
            0.6,
            id="with NaN",
        ),
    ],
)
def test_sum_of_abs_residuals(simulated, target, weight, exp_result):
    """Test function 'sum_of_abs_residuals."""
    result = sum_of_abs_residuals(simulated=simulated, target=target, weighting=weight)

    assert result == pytest.approx(exp_result)


@pytest.mark.parametrize(
    "simulated, target, weight, exp_result",
    [
        pytest.param(
            np.array([1, 2, 3, 4], dtype=float),
            np.array([1.1, 1.8, 3.3, 4.0], dtype=float),
            1,
            0.14,
            id="normal",
        ),
        pytest.param(
            np.array([1, 2, 3, 4, np.nan], dtype=float),
            np.array([1.1, 1.8, 3.3, 4.0, 5], dtype=float),
            1,
            0.14,
            id="with NaN",
        ),
    ],
)
def test_sum_of_squared_residuals(simulated, target, weight, exp_result):
    """Test function 'sum_of_squared_residuals'."""
    result = sum_of_squared_residuals(
        simulated=simulated, target=target, weighting=weight
    )

    assert result == pytest.approx(exp_result)


@pytest.mark.parametrize(
    "simulated, target, weight, free_parameters, exp_result",
    [
        pytest.param(
            np.array([1, 2, 3, 4], dtype=float),
            np.array([1.1, 1.8, 3.3, 4.0], dtype=float),
            1,
            2,
            0.07,
            id="normal",
        ),
        pytest.param(
            np.array([1, 2, 3, 4, np.nan], dtype=float),
            np.array([1.1, 1.8, 3.3, 4.0, 5], dtype=float),
            1,
            2,
            0.07,
            id="with NaN",
        ),
    ],
)
def test_reduced_chi_squared(simulated, target, weight, free_parameters, exp_result):
    """Test function 'reduced_chi_squared'."""
    result = reduced_chi_squared(
        simulated=simulated,
        target=target,
        free_parameters=free_parameters,
        weighting=weight,
    )

    assert result == pytest.approx(exp_result)



# Copyright (c) 2023 Arthur Kadela and Joonas Viuho, Niels Bohr Institute, University of Copenhagen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from pyxel.detectors import CCD
from pyxel.models.charge_transfer import multiplication_register_cic


def test_multiplication_register_cic(ccd_10x10: CCD):
    detector = ccd_10x10

    multiplication_register_cic(
        detector=detector, total_gain=0, gain_elements=1, pcic_rate=0.0, scic_rate=0.0
    )


@pytest.mark.parametrize(
    "total_gain,gain_elements,pcic_rate,scic_rate",
    [
        (-1, 10, 0.0, 0.0),
        (1, -1, -1, 1),
        (-1, -1, 0.005, -0.02),
    ],
)
def test_multiplication_register_cic_bad_inputs(
    ccd_10x10: CCD, total_gain, gain_elements, pcic_rate, scic_rate
):
    detector = ccd_10x10

    with pytest.raises(ValueError, match="Wrong input parameter"):
        multiplication_register_cic(
            detector=detector,
            total_gain=total_gain,
            gain_elements=gain_elements,
            pcic_rate=pcic_rate,
            scic_rate=scic_rate,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from pathlib import Path

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Detector,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_generation import load_charge


@pytest.fixture
def ccd_10x1() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=1,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def profile_10x1() -> np.ndarray:
    """Create a profile of 10x1."""
    data_1d = np.array([100, 100, 100, 100, 100, 0, 0, 0, 0, 0])

    return data_1d


@pytest.fixture
def profile_10x1_txt_filename(profile_10x1: np.ndarray, tmp_path: Path) -> Path:
    """Create a filename with a profile of 10x1."""
    filename = tmp_path / "profile_10x1.txt"
    np.savetxt(fname=filename, X=profile_10x1)

    return filename


@pytest.fixture
def ccd_10x3() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def profile_10x3() -> np.ndarray:
    """Create a profile of 10x3."""
    data_2d = np.zeros(shape=(10, 3))
    data_2d[:5, 0] = 100
    data_2d[:5, 1] = 200
    data_2d[:5, 2] = 300

    data_2d = np.array(
        [
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
    )

    return data_2d


@pytest.fixture
def profile_10x3_txt_filename(profile_10x3: np.ndarray, tmp_path: Path) -> Path:
    """Create a filename with a profile of 10x3."""
    filename = tmp_path / "profile_10x3.txt"
    np.savetxt(fname=filename, X=profile_10x3)

    return filename


def test_charge_profile_10x1(ccd_10x1: CCD, profile_10x1_txt_filename: Path) -> None:
    """Test function 'charge_profile'."""
    detector: Detector = ccd_10x1

    # Check initial detector object
    assert detector.geometry.row == 10
    assert detector.geometry.col == 1
    assert detector.charge.frame.empty

    # Run model
    load_charge(
        detector=detector,
        filename=profile_10x1_txt_filename,
    )

    # Check modified charges in 'detector'
    exp_charges = np.array([[100], [100], [100], [100], [100], [0], [0], [0], [0], [0]])
    charges = detector.charge.array

    np.testing.assert_array_almost_equal(charges, exp_charges)


def test_charge_profile_10x3(ccd_10x3: CCD, profile_10x3_txt_filename: Path) -> None:
    """Test function 'charge_profile'."""
    detector: Detector = ccd_10x3

    # Check initial detector object
    assert detector.geometry.row == 10
    assert detector.geometry.col == 3
    assert detector.charge.frame.empty

    # Run model
    load_charge(
        detector=detector,
        filename=profile_10x3_txt_filename,
    )

    # Check modified charges in 'detector'
    exp_charges = np.array(
        [
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [100, 200, 300],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
    )
    charges = detector.charge.array

    np.testing.assert_array_almost_equal(charges, exp_charges)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""General purposes models or routines for all Group Models."""

from pathlib import Path
from typing import Union

from pyxel.detectors import Detector


def load_detector(detector: Detector, filename: Union[str, Path]) -> None:
    """Load a new detector from a file.

    Raises
    ------
    TypeError
        If the loaded detector has not the same type of the current detector.
    """
    new_detector = Detector.load(filename)

    # Check type of 'new_detector'
    if type(detector) is not type(new_detector):
        raise TypeError(
            f"Wrong detector type from 'filename':'{filename}'. "
            f"Got type: '{type(detector).__name__}', expected '{type(new_detector).__name__}'"
        )

    detector = new_detector


def save_detector(detector: Detector, filename: Union[str, Path]) -> None:
    """Save the current detector into a file."""
    detector.save(filename)
